from aeronaut.request.cloud.v0_9.server_action import ServerActionRequest


class AddStorageToServer(ServerActionRequest):

    def params(self):
        p = super(AddStorageToServer, self).params().copy()
        p.update({
            'server_id': {
                'required': True
            },

            'size_gb': {
                'required': True
            },

            'disk_speed_id': {
                'required': False
            }
        })

        return p

    def url(self):
        u = super(AddStorageToServer, self).url('addLocalStorage')

        u += '&amount={}'.format(self.get_param('size_gb'))

        if self.has_param('disk_speed_id'):
            u += '&speed={}'.format(self.get_param('disk_speed_id'))

        return u



from aeronaut.resource.cloud.resource import Resource, ResourceList


class AdditionalDisk(Resource):

    def _root_(self):
        return "additionalDisk"

    def _members_(self):
        return {
            "id": {
                "xpath": "./*[local-name()='id']"},

            "scsi_id": {
                "xpath": "./*[local-name()='scsiId']",
                "type": int},

            "size_gb": {
                "xpath": "./*[local-name()='diskSizeGb']",
                "type": int},

            "state": {
                "xpath": "./*[local-name()='state']"},
        }


class AdditionalDiskList(ResourceList):

    def _root_(self):
        return "ServerImageWithState"

    def _items_(self):
        return {
            "xpath": "./*[local-name()='additionalDisk']",
            "type": AdditionalDisk
        }


class Disk(Resource):

    def _root_(self):
        return "disk"

    def _members_(self):
        return {
            "id": {
                "xpath": "./@*[local-name()='id']"},

            "scsi_id": {
                "xpath": "./@*[local-name()='scsiId']",
                "type": int},

            "size_gb": {
                "xpath": "./@*[local-name()='sizeGb']",
                "type": int},

            "speed": {
                "xpath": "./@*[local-name()='speed']"},

            "state": {
                "xpath": "./@*[local-name()='state']"},
        }


class DiskList(ResourceList):

    def _root_(self):
        return [
            "image",
            "server",
            "ServerImageWithState",
        ]

    def _items_(self):
        return {
            "xpath": "./*[local-name()='disk']",
            "type": Disk
        }



from aeronaut.resource.cloud.resource import Resource, ResourceList


class Backup(Resource):

    def _root_(self):
        return "backup"

    def _members_(self):
        return {
            "type": {
                "xpath": "./@*[local-name()='type']"},

            "maintenance_status": {
                "xpath": "./@*[local-name()='maintenanceStatus']"}
        }


class DataCenter(Resource):

    def _root_(self):
        return "datacenter"

    def _members_(self):
        return {
            "backup": {
                "xpath": "./*[local-name()='backup']",
                "type": Backup},

            "city": {
                "xpath": "./*[local-name()='city']"},

            "country": {
                "xpath": "./*[local-name()='country']"},

            "display_name": {
                "xpath": "./*[local-name()='displayName']"},

            "hypervisor": {
                "xpath": "./*[local-name()='hypervisor']",
                "type": Hypervisor},

            "is_default": {
                "xpath": "./@*[local-name()='default']",
                "type": bool},

            "location": {
                "xpath": "./@*[local-name()='location']"},

            "networking": {
                "xpath": "./*[local-name()='networking']",
                "type": Networking},

            "state": {
                "xpath": "./*[local-name()='state']"},

            "vpn_url": {
                "xpath": "./*[local-name()='vpnUrl']"},
        }


class DataCenterList(ResourceList):

    def _root_(self):
        return "DatacentersWithMaintenanceStatus"

    def _items_(self):
        return {
            "xpath": "./*[local-name()='datacenter']",
            "type": DataCenter
        }


class DiskSpeed(Resource):

    def _root_(self):
        return "diskSpeed"

    def _members_(self):
        return {
            "abbreviation": {
                "xpath": "./*[local-name()='abbreviation']"},

            "description": {
                "xpath": "./*[local-name()='description']"},

            "display_name": {
                "xpath": "./*[local-name()='displayName']"},

            "id": {
                "xpath": "./@*[local-name()='id']"},

            "is_available": {
                "xpath": "./@*[local-name()='available']",
                "type": bool},

            "is_default": {
                "xpath": "./@*[local-name()='default']",
                "type": bool},
        }


class DiskSpeedList(ResourceList):

    def _root_(self):
        return "hypervisor"

    def _items_(self):
        return {
            "xpath": "./*[local-name()='diskSpeed']",
            "type": DiskSpeed
        }


class Hypervisor(Resource):

    def _root_(self):
        return "hypervisor"

    def _members_(self):
        return {
            "disk_speeds": {
                "xpath": ".",
                "type": DiskSpeedList},

            "maintenance_status": {
                "xpath": "./@*[local-name()='maintenanceStatus']"},

            "min_disk_size_gb": {
                "xpath": "./*[local-name()='property']"
                         "[@name='MIN_DISK_SIZE_GB']/@value",
                "type": int},

            "type": {
                "xpath": "./@*[local-name()='type']"}
        }


class Networking(Resource):

    def _root_(self):
        return "networking"

    def _members_(self):
        return {
            "maintenance_status": {
                "xpath": "./@*[local-name()='maintenanceStatus']"},

            "max_server_to_vip_connections": {
                "xpath": "./*[local-name()='property']"
                         "[@name='MAX_SERVER_TO_VIP_CONNECTIONS']/@value",
                "type": int},

            "type": {
                "xpath": "./@*[local-name()='type']",
                "type": int}
        }



from aeronaut.resource.cloud.resource import Resource, ResourceList


class ActionStatus(Resource):

    def _root_(self):
        return "status"

    def _members_(self):
        return {
            "action": {
                "xpath": "./*[local-name()='action']"},

            "failure_reason": {
                "xpath": "./*[local-name()='failureReason']"},

            "request_time": {
                "xpath": "./*[local-name()='requestTime']"},

            "steps": {
                "xpath": ".",
                "type": ActionStatusStepsList},

            "update_time": {
                "xpath": "./*[local-name()='updateTime']"},

            "username": {
                "xpath": "./*[local-name()='userName']"}
        }


class ActionStatusStep(Resource):

    def _root_(self):
        return "step"

    def _members_(self):
        return {
            "name": {
                "xpath": "./*[local-name()='name']"},

            "number": {
                "xpath": "./*[local-name()='number']",
                "type": int},

            "percent_complete": {
                "xpath": "./*[local-name()='percentComplete']",
                "type": int}
        }


class ActionStatusStepsList(ResourceList):

    def _root_(self):
        return "status"

    def _items_(self):
        return {
            "xpath": "./*[local-name()='step']",
            "type": ActionStatusStep
        }



from aeronaut.request.cloud.v0_9.request import Request


class GetMyAccount(Request):

    def http_method(self):
        return 'get'

    def url(self):
        return '{base_url}/myaccount'.format(base_url=self.base_url)



from os.path import expanduser, exists, join
import re
import requests
import yaml

from aeronaut.resource.cloud.resource import RootElementNotFoundError
from aircraft.response import Response


# ==========
# EXCEPTIONS
# ==========

class AuthenticationError(Exception):

    def __init__(self, response):
        message = 'Can\'t authenticate against {remote}. ' \
                  'Server returned HTTP status {code}'.format(
                      remote=response.url, code=response.status_code)
        super(AuthenticationError, self).__init__(message)

        self.response = response


class NotAuthenticatedError(Exception):

    def __init__(self, response):
        message = 'The operation requires authentication beforehand'
        super(AuthenticationError, self).__init__(message)


class OperationForbiddenError(Exception):

    def __init__(self, status):
        message = "The server responded with a 403: Forbidden error. " \
                  "{code}: {details}".format(code=status.result_code,
                                             details=status.result_detail)
        super(OperationForbiddenError, self).__init__(message)

        self.status = status


class UnauthorizedError(Exception):

    def __init__(self):
        message = "Unable to complete the request due to authorization " \
                  "error. The session has likely expired. Please check " \
                  "and re-authenticate as needed."
        super(UnauthorizedError, self).__init__(message)


# ================
# CONNECTION CLASS
# ================

class CloudConnection(object):
    """Represents a connection to a DiData Cloud provider"""

    def __init__(self, endpoint, api_version='v0.9'):
        """
        :param str endpoint: The endpoint to connect to when
            :meth:`~aeronaut.cloud.CloudConnection.authenticate` is called.

        :param str api_version: Optional. The API version to use.
            Defaults to v0.9 when not provided. Note that the default may
            be updated to a later API version in future releases.
        """
        self._endpoint = endpoint
        self._api_version = api_version
        self._my_account = None

    # =================
    # PUBLIC PROPERTIES
    # =================

    @property
    def api_version(self):
        return self._api_version

    @property
    def base_url(self):
        return 'https://{}'.format(self.endpoint)

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def is_authenticated(self):
        return self.my_account is not None

    @property
    def my_account(self):
        return self._my_account

    # ==================
    # PRIVATE PROPERTIES
    # ==================

    @property
    def _dotfile_credentials(self):
        """
        Gets the applicable credentials from ~/.aeronaut when available. If
        not available, None will be returned.
        """
        path = join(expanduser('~'), '.aeronaut')

        if not exists(path):
            return None

        if not hasattr(self, '__dotfile_credentials__'):
            with open(path, 'r') as dotfile:
                d = yaml.load(dotfile)
                if self.endpoint not in d:
                    return None
                self.__dotfile_credentials__ = d[self.endpoint]

        return self.__dotfile_credentials__

    @property
    def _http_session(self):
        if not hasattr(self, '__http_session__'):
            self.__http_session__ = requests.Session()

        return self.__http_session__

    # ==============
    # PUBLIC METHODS
    # ==============

    def add_storage_to_server(self, server_id, size_gb, disk_speed_id=None,
                              org_id=None):
        """Creates an additional disk/volume for an existing server

        :param str server_id: The ID of the server to add storage to.

        :param int size_gb: The size of the disk in GB.

        :param str disk_speed_id: Optional. The id of the disk speed to use. This
            can be obtained via :meth:`aeronaut.resource.cloud.data_center.DataCenter.hypervisor.disk_speeds`.
            See the sample usage below for guidance.

        :param str org_id: Optional. The organization ID that owns the
            server. Defaults to the current user's :py:attr:`~aeronaut.resource.cloud.account.Account.org_id`
            if not provided.

        :returns: The status of the operation.
        :rtype: :class:`aeronaut.resource.cloud.server.AddStorageToServerStatus`

        Sample usage:

        .. code-block:: python

            import aeronaut.cloud

            conn = aeronaut.cloud.connect('api-na.dimensiondata.com')

            filters = [
                ['name', '==', 'myserver01']
            ]

            servers = conn.list_servers(filters=filters)
            server = servers[0]

            dc = next(dc for dc in conn.list_data_centers() if dc.is_default)
            speed = dc.hypervisor.disk_speeds[0]

            # The disk_speed_id argument below is optional
            result = conn.add_storage_to_server(server_id=server.id,
                                                size_gb=13,
                                                disk_speed_id=speed.id)

            print result.is_success
            print result.result_detail
        """  # NOQA
        params = {
            'org_id': self._ensure_org_id(org_id),
            'server_id': server_id,
            'size_gb': size_gb,
            'disk_speed_id': disk_speed_id
        }
        response = self.request('add_storage_to_server', params=params)
        self._raise_if_unauthorized(response)
        return self._deserialize('server.AddStorageToServerStatus',
                                 response.body)

    def authenticate(self):
        """Authenticates against the endpoint provided during initialization

        To successfully authenticate against the given endpoint, the method
        will look for a file in the path ``~/.aeronaut`` with a section name
        matching the same endpoint. The file must be valid YAML with the
        following structure::

            api-na.dimensiondata.com:
                username: myuser
                password: mypassword

            another.endpoint.com:
                username: myotheruser
                password: mypasswordtoo!

        If the backend server does not return an HTTP status code of 200 or
        201, the method will raise
        :class:`aeronaut.cloud.AuthenticationError`
        """
        kwargs = {}

        if self._dotfile_credentials is not None:
            credentials = self._dotfile_credentials
            username = credentials['username']
            password = credentials['password']
            kwargs['auth'] = (username, password)

        # The server does pre-emptive basic authentication and self.request()
        # Provides basic auth data when available. So calling any method will
        # automatically authenticate us. We just happen to choose the
        # 'get_my_account' request for this process so that we will have a copy
        # of the account in memory.
        response = self.request('get_my_account', **kwargs)

        if response.status_code in [200, 201]:
            self._my_account = self._deserialize('my_account.MyAccount',
                                                 response.body)
        else:
            raise AuthenticationError(response)

    def clean_failed_server_deployment(self, server_id, org_id=None):
        """Removes a failed server deployment from the list of pending
        deployed servers

        :param str server_id: The id of the server you wish to remove.

        :param str org_id: Optional. The organization ID that owns the
            server. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :returns: The status of the operation
        :rtype: :class:`aeronaut.resource.cloud.server.CleanFailedServerDeploymentStatus`
        """  # NOQA
        params = {
            'org_id': self._ensure_org_id(org_id),

            'server_id': server_id
        }
        response = self.request('clean_failed_server_deployment',
                                params=params)
        self._raise_if_unauthorized(response)
        return self._deserialize('server.CleanFailedServerDeploymentStatus',
                                 response.body)

    def create_acl_rule(self, network_id, name, position, action, protocol,
                        type, source_ip=None, source_netmask=None,
                        dest_ip=None, dest_netmask=None, from_port=None,
                        to_port=None, org_id=None):
        """Creates an ACL Rule

        :param str network_id: The id of the network where the ACL rule
            is to be created.

        :param str name: The name of the ACL rule.

        :param int position: The position of the ACL rule. Valid range
            is 100-500 inclusive.

        :param str action: One of PERMIT or DENY.

        :param str protocol: One of IP, ICMP, TCP, UDP.

        :param str type: For inbound rules, provide OUTSIDE_ACL. For
            outbound rules, provide INSIDE_ACL.

        :param str source_ip: Optional. IP address of the traffic
            source.

        :param str source_netmaks: Optional. Netmask for the source IP.

        :param str dest_ip: Optional. IP address of the traffic
            destination.

        :param str dest_netmask: Optional. for the destination IP.

        :param int from_port: Optional. Valid range is 1-65535
            inclusive

        :param int to_port: Optional. Valid range is 1-65535 inclusive

        :param str org_id: Optional. The organization ID that owns the
            network. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :returns: The status of the operation
        :rtype: :class:`aeronaut.resource.cloud.acl.CreateAclRuleStatus`
        """
        org_id = self._ensure_org_id(org_id)

        params = {
            'type': type,
            'action': action,
            'dest_ip': dest_ip,
            'dest_netmask': dest_netmask,
            'from_port': from_port,
            'name': name,
            'network_id': network_id,
            'org_id': org_id,
            'position': position,
            'protocol': protocol,
            'source_ip': source_ip,
            'source_netmask': source_netmask,
            'to_port': to_port
        }

        response = self.request('create_acl_rule', params=params)
        self._raise_if_unauthorized(response)

        # The underlying REST API is inconsistent. When succesful, it returns
        # an AclRule element, otherwise it returns a Status element. This is an
        # attempt to make our own return value consistent.
        try:
            return self._deserialize('acl.CreateAclRuleStatus', response.body)
        except RootElementNotFoundError:
            xml = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
            <ns6:Status xmlns:ns6="http://oec.api.opsource.net/schemas/general">
                <ns6:operation>Add ACL Rule</ns6:operation>
                <ns6:result>SUCCESS</ns6:result>
                <ns6:resultDetail>ACL Rule created succesfully</ns6:resultDetail>
                <ns6:resultCode>REASON_0</ns6:resultCode>
            </ns6:Status>
            """  # NOQA
            return self._deserialize('acl.CreateAclRuleStatus', xml)

    def create_network(self, location, name, org_id=None, description=None):
        """Creates a network with the given name in the given location

        :param str location: The data center where the network is to be
            created.

        :param str name: The name of the network

        :param str org_id: Optional. The organization ID that owns the
            network. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :returns: The status of the operation
        :rtype: :class:`aeronaut.resource.cloud.network.CreateNetworkStatus`
        """
        org_id = self._ensure_org_id(org_id)

        params = {
            'org_id': org_id,
            'location': location,
            'name': name
        }

        if description:
            params['description'] = description

        response = self.request('create_network', params=params)
        self._raise_if_unauthorized(response)

        return self._deserialize('network.CreateNetworkStatus', response.body)

    def delete_acl_rule(self, network_id, rule_id, org_id=None):
        """Deletes the given rule from the given network

        :param str network_id: The ID of the network associated with
            the ACL rule.

        :param str rule_id: The ID of the ACL rule.

        :param str org_id: Optional. The organization ID that owns the
            network. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :returns: The status of the operation
        :rtype: :class:`aeronaut.resource.cloud.acl.DeleteAclRuleStatus`
        """
        org_id = self._ensure_org_id(org_id)

        params = {
            'org_id': org_id,
            'network_id': network_id,
            'rule_id': rule_id
        }

        response = self.request('delete_acl_rule', params=params)
        self._raise_if_unauthorized(response)

        return self._deserialize('acl.DeleteAclRuleStatus', response.body)

    def delete_server(self, server_id, org_id=None):
        """Delete the given server. Note that a Server must be stopped before
        it can be deleted

        :param str server_id: The ID of the server to delete.

        :param str org_id: Optional. The organization ID that owns the
            server. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :returns: The status of the operation
        :rtype: :mod:`aeronaut.resource.cloud.server.DeleteServerStatus`
        """
        params = {
            'org_id': self._ensure_org_id(org_id),
            'server_id': server_id
        }

        response = self.request('delete_server', params=params)
        self._raise_if_unauthorized(response)
        return self._deserialize('server.DeleteServerStatus', response.body)

    def deploy_server(self, name, image_id, org_id=None, **kwargs):
        """Deploys a new server from an existing customer or base image

        :param str name: Human-friendly name of the server.

        :param str image_id: ID of the image to use for this server.

        :param str org_id: Optional. The organization ID that owns the
            server. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided. authenticated user's account will be used.

        Additional arguments are listed in the source code of
        :class:`aeronaut.request.cloud.v0_9.deploy_server.DeployServer`

        :returns: The status of the operation
        :rtype: :class:`aeronaut.resource.cloud.server.DeployServerStatus`

        Sample usage:

        .. code-block:: python

            import aeronaut.cloud

            conn = aeronaut.cloud.connect('api-na.dimensiondata.com')

            dcs = conn.list_data_centers()
            dc = next(dc for dc in dcs if dc.is_default)

            images = conn.list_base_images(filters=filters)
            image = images[0]

            networks = conn.list_networks(location=dc.location)
            network = networks[0]

            status = conn.deploy_server(name='My new server',
                                        description='A new deployment',
                                        image_id=image.id,
                                        start=True,
                                        admin_password='asdfkjhw',
                                        network_id=network.id)
            print status.is_success

        Note how there is no argument for adding disks above. To add one,
        you need to call :meth:`~aeronaut.cloud.CloudConnection.add_storage_to_server`
        """  # NOQA
        kwargs['org_id'] = self._ensure_org_id(org_id)
        kwargs['name'] = name
        kwargs['image_id'] = image_id

        response = self.request('deploy_server', params=kwargs)
        self._raise_if_unauthorized(response)

        return self._deserialize('server.DeployServerStatus', response.body)

    def does_image_name_exist(self, image_name, location, org_id=None):
        """Returns True if the given image name exists in the data center

        :param str image_name: The name of the image

        :param str location: Location where the image is expected to exist.
            This can be taken from the `location` property of the
            :class:`~aeronaut.resource.cloud.data_center.DataCenter` resource
            object.

        :param str org_id: Optional. The organization ID that owns the
            image. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :returns: :class:`True` if the image_name exists. :class:`False` if
            otherwise.
        :rtype: :class:`bool`
        """
        org_id = self._ensure_org_id(org_id)

        params = {
            'org_id': org_id,
            'image_name': image_name,
            'location': location
        }

        response = self.request('does_image_name_exist', params=params)
        self._raise_if_unauthorized(response)
        exists = self._deserialize('image.ImageNameExists', response.body)

        return exists.is_true

    def get_server_image(self, image_id, org_id=None):
        """Returns a full description of the indicated server image

        :param str image_id: The id of the image to retrieve

        :param str org_id: Optional. The organization ID that owns the
            image. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :returns: The image with the given ID
        :rtype: :class:`aeronaut.resource.cloud.image.Image`
        """
        params = {
            'org_id': self._ensure_org_id(org_id),
            'image_id': image_id
        }

        response = self.request('get_server_image', params=params)
        self._raise_if_unauthorized(response)
        image = self._deserialize('image.ServerImage', response.body)

        return image

    def list_acl_rules(self, network_id, org_id=None):
        """Returns a list of ACL rules

        :param str network_id: The ID of the network whose ACL rules you want
            to retrieve.

        :param str org_id: Optional. The organization ID that owns the
            network. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :returns: A list of ACL rules in the given network
        :rtype: :class:`aeronaut.resource.cloud.acl.AclRuleList`
        """
        if not org_id:
            org_id = self.my_account.org_id

        params = {
            'network_id': network_id,
            'org_id': org_id
        }

        response = self.request('list_acl_rules', params=params)
        self._raise_if_unauthorized(response)

        return self._deserialize('acl.AclRuleList', response.body)

    def list_base_images(self, filters=None, page_size=None, page_number=None,
                         sort=None):
        """Returns a list of base images

        :param list filters: A list of 3-element lists specifying the filters
            to apply when retrieving the base images.

        :param int page_size: The maximum number of items per page

        :param int page_number: The page number to return

        :param list sort: A list of strings specifying the order of the list.

        :rtype: :class:`aeronaut.resource.cloud.image.ImageList`

        Here is an example usage with the above parameters included:

        .. code-block:: python

            filters = [
                ["location", "==", "NA5"],
                ["location", "==", "NA3"],
                ["name", "like", "Windows*"]
            ]

            page_size = 5
            page_number = 1

            sort = [
                "location ASC",
                "name DESC"
            ]

            images = conn.list_base_images(filters=filters,
                                           sort=sort,
                                           page_size=page_size,
                                           page_number=page_number)

            print images.page_number, "of", images.total_pages
            print images.page_size
            print images.total_count
            print len(images)
            for image in images:
                print image.id, image.name
        """
        return self.list_images('base',
                                filters=filters,
                                page_size=page_size,
                                page_number=page_number,
                                sort=sort)

    def list_customer_images(self, org_id=None, filters=None, page_size=None,
                             page_number=None, sort=None):
        """Returns a list of customer images

        :rtype: :class:`aeronaut.resource.cloud.image.ImageList`
        """
        return self.list_images(self._ensure_org_id(org_id),
                                filters=filters,
                                page_size=page_size,
                                page_number=page_number,
                                sort=sort)

    def list_data_centers(self, org_id=None, filters=None, page_size=None,
                          page_number=None, sort=None):
        """Returns a list of data centers

        :rtype: :class:`aeronaut.resource.cloud.data_center.DataCenterList`
        """
        params = {
            'org_id': self._ensure_org_id(org_id),
            'filters': filters,
            'page_size': page_size,
            'page_number': page_number,
            'sort': sort
        }
        response = self.request('list_data_centers', params=params)
        self._raise_if_unauthorized(response)
        return self._deserialize('data_center.DataCenterList', response.body)

    def list_images(self, base_or_org_id, **kwargs):
        """Returns a list of base or customer images. While this is available,
        you are better off calling either
        :meth:`~aeronaut.cloud.CloudConnection.list_customer_images` or
        :meth:`~aeronaut.cloud.CloudConnection.list_base_images`

        :rtype: :class:`aeronaut.resource.cloud.image.ImageList`
        """
        kwargs['base_or_org_id'] = base_or_org_id

        response = self.request('list_images', params=kwargs)
        self._raise_if_unauthorized(response)
        images = self._deserialize('image.ImageList', response.body)

        return images

    def list_networks(self, org_id=None, location=None):
        """Returns a list of networks

        :param str org_id: Optional. The organization ID that owns the
            networks. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :param str location: Location where the image is expected to exist.
            This can be taken from the `location` property of the
            :class:`~aeronaut.resource.cloud.data_center.DataCenter` resource
            object.

        :rtype: :class:`aeronaut.resource.cloud.network.NetworkList`
        """
        org_id = self._ensure_org_id(org_id)

        params = {'org_id': org_id}

        if isinstance(location, str):
            params['location'] = location

        response = self.request('list_networks', params=params)
        self._raise_if_unauthorized(response)

        return self._deserialize('network.NetworkList', response.body)

    def list_servers(self, org_id=None, filters=None, page_size=None,
                     page_number=None, sort=None):
        """Returns a list of servers

        :param str org_id: Optional. The organization ID that owns the
            server. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :rtype: :class:`aeronaut.resource.cloud.server.ServerList`
        """
        params = {
            'org_id': self._ensure_org_id(org_id),
            'filters': filters,
            'page_size': page_size,
            'page_number': page_number,
            'sort': sort
        }

        response = self.request('list_servers', params=params)
        self._raise_if_unauthorized(response)
        return self._deserialize('server.ServerList', response.body)

    def modify_server(self, server_id, name=None, description=None,
                      cpu_count=None, memory=None, org_id=None):
        """Changes the name, description, CPU, or RAM profile of an existing
        deployed server.

        :param str org_id: Optional. The organization ID that owns the
            server. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :param str name: Optional. New name of the server.

        :param str description: Optional. New description of the server.

        :param int cpu_count: Optional. New number of virtual CPUs.

        :param int memory: Optional. New total memory in MB. Value must be a
                multiple of 1024.

        :param str server_id: The ID of the server to modify.

        :rtype: :class:`aeronaut.resource.cloud.server.ModifyServerStatus`
        """
        params = {
            'org_id': self._ensure_org_id(org_id),
            'server_id': server_id
        }

        for var in ['name', 'description', 'cpu_count', 'memory']:
            if eval(var):
                params[var] = eval(var)

        response = self.request('modify_server', params=params)
        self._raise_if_unauthorized(response)
        return self._deserialize('server.ModifyServerStatus', response.body)

    def poweroff_server(self, server_id, org_id=None):
        """Forcefully power off the server

        :param str server_id: The ID of the server you want to power off.

        :param str org_id: Optional. The organization ID that owns the
            server. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :rtype: :class:`aeronaut.resource.cloud.server.PoweroffServerStatus`
        """
        params = {
            'org_id': self._ensure_org_id(org_id),
            'server_id': server_id
        }

        response = self.request('poweroff_server', params=params)
        self._raise_if_unauthorized(response)
        return self._deserialize('server.PoweroffServerStatus', response.body)

    def reboot_server(self, server_id, org_id=None):
        """Gracefully reboot a server

        :param str server_id: The ID of the server you want to power off.

        :param str org_id: Optional. The organization ID that owns the
            server. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :rtype: :class:`aeronaut.resource.cloud.server.RebootServerStatus`
        """
        params = {
            'org_id': self._ensure_org_id(org_id),
            'server_id': server_id
        }

        response = self.request('reboot_server', params=params)
        self._raise_if_unauthorized(response)
        return self._deserialize('server.RebootServerStatus', response.body)

    def reset_server(self, server_id, org_id=None):
        """Forcefully power cycle the server.

        :param str server_id: The ID of the server you want to power off.

        :param str org_id: Optional. The organization ID that owns the
            server. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :rtype: :class:`aeronaut.resource.cloud.server.ResetServerStatus`
        """
        params = {
            'org_id': self._ensure_org_id(org_id),
            'server_id': server_id
        }

        response = self.request('reset_server', params=params)
        self._raise_if_unauthorized(response)
        return self._deserialize('server.ResetServerStatus', response.body)

    def shutdown_server(self, server_id, org_id=None):
        """Gracefully stop a server

        :param str server_id: The ID of the server you want to power off.

        :param str org_id: Optional. The organization ID that owns the
            server. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :rtype: :class:`aeronaut.resource.cloud.server.ShutdownServerStatus`
        """
        params = {
            'org_id': self._ensure_org_id(org_id),
            'server_id': server_id
        }

        response = self.request('shutdown_server', params=params)
        self._raise_if_unauthorized(response)
        return self._deserialize('server.ShutdownServerStatus', response.body)

    def start_server(self, server_id, org_id=None):
        """Powers on an existing server

        :param str server_id: The ID of the server you want to power off.

        :param str org_id: Optional. The organization ID that owns the
            server. Defaults to the current user's
            :py:attr:`~aeronaut.resource.cloud.account.Account.org_id` if not
            provided.

        :rtype: :class:`aeronaut.resource.cloud.server.StartServerStatus`
        """
        params = {
            'org_id': self._ensure_org_id(org_id),
            'server_id': server_id
        }

        response = self.request('start_server', params=params)
        self._raise_if_unauthorized(response)
        return self._deserialize('server.StartServerStatus', response.body)

    def request(self, req_name, api_version=None, params={}, auth=None):
        """Sends a request to the provider. This is a low-level method used by
        all other operations exposed by this class. You do not need to call
        it directly but it's here anyway in case you really want to use it.

        :param str req_name: The name of the request. For a list of available
            requests, see :mod:`aeronaut.request.cloud`

        :param str api_version: The API version to use.

        :param dict params: Parameters to be supplied to the request. This can
            vary between requests.

        :param tuple auth: A tuple of two strings, username and password, to
            use should the server require authentication.

        Example:

        .. code-block:: python

            params = {
                'cpu_count': 2,
                'memory': 2048
            }

            response = self.request('modify_server',
                                    api_version='v0.9',
                                    params=params)
            print response.status_code

        The above call results in the class
        :class:`aeronaut.request.cloud.v0_9.modify_server.ModifyServer`
        being used to send a request to the provider.

        * The version ``v0_9`` is inferred from the ``api_version`` argument
        * The class ``ModifyServer`` is inferred from the first argument,
          ``modify_server``
        * What values go into the params argument is determined by whatever
          the :meth:`~aeronaut.cloud.request.v0_9.modify_server.ModifyServer.params`
          method of the request class returns.

        Note that this method returns a raw response object, not an object
        representation of the XML returned by the server.
        """  # NOQA
        if api_version is None:
            api_version = self.api_version

        request = self._build_request(req_name, api_version, params)
        kwargs = {}

        if auth is not None:
            kwargs['auth'] = auth

        if request.http_method() in ['post', 'put'] \
                and request.body() is not None:
            kwargs['data'] = request.body()

        if request.headers():
            kwargs['headers'] = request.headers()

        http_method = getattr(self._http_session, request.http_method())
        return Response(http_method(request.build_url(), **kwargs), request)

    # ===============
    # PRIVATE METHODS
    # ===============

    def _build_request(self, req_name, api_version, params):
        classname = ''.join([s.capitalize() for s in req_name.split('_')])
        api_version = re.sub('\.', '_', api_version)
        modname = 'aeronaut.request.cloud.{api_version}.{req_name}' \
                  .format(api_version=api_version,
                          req_name=req_name)

        try:
            mod = __import__(modname, fromlist=[classname])
        except ImportError:
            raise ImportError("no module named {}.{}"
                              .format(modname, classname))

        klass = getattr(mod, classname)
        return klass(base_url=self.base_url, params=params)

    def _deserialize(self, classname, xml):
        modname, classname = classname.split('.')
        modname = 'aeronaut.resource.cloud.{}'.format(modname)

        try:
            mod = __import__(modname, fromlist=[classname])
        except ImportError:
            raise ImportError("Cannot import module named {}.{}"
                              .format(modname, classname))

        klass = getattr(mod, classname)
        return klass(xml=xml)

    def _ensure_org_id(self, org_id):
        if org_id:
            return org_id
        else:
            if not self.my_account:
                raise NotAuthenticatedError()

            return self.my_account.org_id

    def _raise_if_unauthorized(self, response):
        if response.status_code == 401:
            raise UnauthorizedError()
        elif response.status_code == 403:
            status = self._deserialize('resource.Status', response.body)
            raise OperationForbiddenError(status)


def connect(endpoint):
    """Creates a new connection to a DiData Cloud Server or DiData Cloud
    Network provider. Multiple calls to :py:func:`~aeronaut.cloud.connect` can
    be made to establish connections to multiple providers.

    :param str endpoint: The endpoint to connect to. For example:
        ``api-na.dimensiondata.com``

    :return: An authenticated connection to the cloud provider
    :rtype: :class:`~aeronaut.cloud.CloudConnection`

    Here is a simple usage example::

        from aeronaut.cloud import connect

        conn = connect('api-na.dimensiondata.com')
        servers = aeronaut.list_servers()

    To successfully authenticate against the given endpoint, the function will
    look for a file in the path ``~/.aeronaut`` with a section name matching
    the same endpoint. The file must be valid YAML with the following
    structure::

        api-na.dimensiondata.com:
            username: myuser
            password: mypassword

        another.endpoint.com:
            username: myotheruser
            password: mypasswordtoo!

    Note that this function is really just syntactic sugar that wraps in a
    single step the act of instantiating a
    :class:`~aeronaut.cloud.CloudConnection` object and calling
    :meth:`~aeronaut.cloud.CloudConnection.authenticate` against it. To learn
    more about the authentication process,
    see :meth:`aeronaut.cloud.CloudConnection.authenticate`.

    Here is a more detailed usage example::

        from aeronaut.cloud import connect

        cloud = connect('api-na.dimensiondata.com')

        images = cloud.list_base_images()
        image = images[0]

        networks = cloud.list_networks()
        network = networks[0]

        status = conn.deploy_server(name='My New Server',
                                    description='First server deployment',
                                    image_id=image.id,
                                    start=True,
                                    admin_password='32456sdkddd',
                                    network_id=network.id)

        print status.is_success

    All requests such as :meth:`~aeronaut.cloud.CloudConnection.deploy_server`
    and :meth:`~aeronaut.cloud.CloudConnection.list_base_images` are delegated
    to request classes under the :mod:`aeronaut.request.cloud` package. Whereas
    the return values are instances of classes under the
    :mod:`aeronaut.resource.cloud` package. Documentation is currently sparse
    in both packages so you are required to read the source code to fully
    understand what the request and resource objects do.
    """
    conn = CloudConnection(endpoint=endpoint)
    conn.authenticate()
    return conn



from aeronaut.request.cloud.v0_9.request import Request


class CreateAclRule(Request):

    def params(self):
        return {
            'type': {
                'required': True
            },
            'action': {
                'required': True
            },
            'dest_ip': {
                'required': False
            },
            'dest_netmask': {
                'required': False
            },
            'from_port': {
                'required': True
            },
            'name': {
                'required': True
            },
            'network_id': {
                'required': True
            },
            'org_id': {
                'required': True
            },
            'position': {
                'required': True
            },
            'protocol': {
                'required': True
            },
            'source_ip': {
                'required': False
            },
            'source_netmask': {
                'required': False
            },
            'to_port': {
                'required': True
            }
        }

    def http_method(self):
        return 'post'

    def url(self):
        template = '{base_url}/{org_id}/network/{network_id}/aclrule'

        return template.format(base_url=self.base_url,
                               network_id=self.get_param('network_id'),
                               org_id=self.get_param('org_id'))

    def body(self):
        template = """
            <AclRule xmlns="http://oec.api.opsource.net/schemas/network">
                <name>{name}</name>
                <position>{position}</position>
                <action>{action}</action>
                <protocol>{protocol}</protocol>
"""

        if self.has_param('source_ip'):
            template += """
                <sourceIpRange>
                    <ipAddress>{source_ip}</ipAddress>
                    <netmask>{source_netmask}</netmask>
                </sourceIpRange>
"""
        else:
            template += """
                <sourceIpRange/>
"""

        if self.has_param('dest_ip'):
            template += """
                <destinationIpRange>
                    <ipAddress>{dest_ip}</ipAddress>
                    <netmask>{dest_netmask}</netmask>
                </destinationIpRange>
"""
        else:
            template += """
                <destinationIpRange/>
"""

        if self.get_param('from_port') == self.get_param('to_port'):
            template += """
                <portRange>
                    <type>EQUAL_TO</type>
                    <port1>{from_port}</port1>
                </portRange>
"""
        else:
            template += """
                <portRange>
                    <type>RANGE</type>
                    <port1>{from_port}</port1>
                    <port2>{to_port}</port2>
                </portRange>
"""

        template += """
                <type>{acl_type}</type>
            </AclRule>"""

        body = template.format(acl_type=self.get_param('type'),
                               action=self.get_param('action'),
                               dest_ip=self.get_param('dest_ip'),
                               dest_netmask=self.get_param('dest_netmask'),
                               from_port=self.get_param('from_port'),
                               name=self.get_param('name'),
                               network_id=self.get_param('network_id'),
                               position=self.get_param('position'),
                               protocol=self.get_param('protocol'),
                               source_ip=self.get_param('source_ip'),
                               source_netmask=self.get_param('source_netmask'),
                               to_port=self.get_param('to_port'))

        return body



import urllib

from aeronaut.request.cloud.v0_9.request import Request


class ModifyServer(Request):

    def params(self):
        return {
            'cpu_count': {
                'required': False,
                'default': None
            },

            'description': {
                'required': False,
                'default': None
            },

            'memory': {
                'required': False,
                'default': None
            },

            'name': {
                'required': False,
                'default': None
            },

            'server_id': {
                'required': True
            },

            'org_id': {
                'required': True
            }
        }

    def http_method(self):
        return 'post'

    def url(self):
        template = "{base_url}/{org_id}/server/{server_id}"

        return template.format(base_url=self.base_url,
                               org_id=self.get_param('org_id'),
                               server_id=self.get_param('server_id'))

    def headers(self):
        headers = super(ModifyServer, self).headers()
        headers['Content-type'] = 'application/x-www-form-urlencoded'
        return headers

    def body(self):
        payload = {}

        for param in ['description', 'memory', 'name']:
            if self.has_param(param):
                payload[param] = self.get_param(param)

        if self.has_param('cpu_count'):
            payload['cpuCount'] = self.get_param('cpu_count')

        return urllib.urlencode(payload)



from aeronaut.resource.cloud.resource import Resource, ResourceList, Status


class CreateNetworkStatus(Status):
    pass


class Network(Resource):

    def _root_(self):
        return 'network'

    def _members_(self):
        return {
            "description": {
                "xpath": "./*[local-name()='description']"},

            "id": {
                "xpath": "./*[local-name()='id']"},

            "is_multicast": {
                "xpath": "./*[local-name()='multicast']",
                "type": bool},

            "location": {
                "xpath": "./*[local-name()='location']"},

            "name": {
                "xpath": "./*[local-name()='name']"},

            "private_net": {
                "xpath": "./*[local-name()='privateNet']"}
        }


class NetworkList(ResourceList):

    def _root_(self):
        return 'NetworkWithLocations'

    def _items_(self):
        return {
            "xpath": "./*[local-name()='network']",
            "type": Network
        }



from aeronaut.request.cloud.v0_9.request import Request


class ListAclRules(Request):

    def params(self):
        return {
            'network_id': {
                'required': True
            },
            'org_id': {
                'required': True
            }
        }

    def http_method(self):
        return 'get'

    def url(self):
        template = '{base_url}/{org_id}/network/{network_id}/aclrule'

        return template.format(base_url=self.base_url,
                               network_id=self.get_param('network_id'),
                               org_id=self.get_param('org_id'))



from aircraft.value_object import ValueObject


# ==========
# EXCEPTIONS
# ==========

class MissingRequiredParametersError(Exception):
    def __init__(self, missing):
        if len(missing) == 1:
            substr = ' is'
        else:
            substr = 's are'

        message = 'The following parameter{} required: {}'.format(
            substr, ', '.join(missing))
        super(MissingRequiredParametersError, self).__init__(message)


class MissingDefaultValueError(Exception):
    """
    If this error is raised, that means the request class has a bug. Double
    check that all of its optional parameters have defaults.
    """
    def __init__(self, missing):
        message = "Optional parameter '{}' does not declare a default value" \
                  .format(missing)
        super(MissingDefaultValueError, self).__init__(message)


class UnexpectedParameterNameError(Exception):
    def __init__(self, name):
        message = 'Unknown parameter {}'.format(name)
        super(UnexpectedParameterNameError, self).__init__(message)


# ============
# PARAMS CLASS
# ============

class Params(ValueObject):

    def __init__(self, declared, actual):
        self._attrs = {}

        required_keys = [k for k, v in declared.items() if v['required']]
        optional_keys = list(set(declared.keys()) - set(required_keys))
        missing_keys = list(set(required_keys) - set(actual.keys()))
        default_keys = list(set(optional_keys) - set(actual.keys()))

        if missing_keys:
            raise MissingRequiredParametersError(missing_keys)

        for key in actual.keys():
            if key not in declared.keys():
                raise UnexpectedParameterNameError(key)

            self._attrs[key] = actual[key]

        for key in default_keys:
            if 'default' not in declared[key]:
                raise MissingDefaultValueError(key)

            self._attrs[key] = declared[key]['default']

    def __getitem__(self, key):
        return self._attrs[key]



from aeronaut.resource.cloud.resource import ResourceList


class SoftwareLabelList(ResourceList):

    def _root_(self):
        return [
            "image",
            "server",
            "ServerImageWithState"
        ]

    def _items_(self):
        return {
            "xpath": "./*[local-name()='softwareLabel']"
        }



from aeronaut.request.cloud.v0_9.request import Request


class UpdateAdminAccount(Request):

    def params(self):
        return {
            "department": {
                "required": False},

            "email": {
                "required": False},

            "first_name": {
                "required": False},

            "full_name": {
                "required": False},

            "last_name": {
                "required": False},

            "org_id": {
                "required": True},

            "password": {
                "required": False},

            "username": {
                "required": True}
        }

    def http_method(self):
        return 'post'

    def url(self):
        return '{base_url}/{org_id}/account/{username}' \
               .format(base_url=self.base_url,
                       org_id=self.get_param('org_id'),
                       username=self.get_param('username'))

    def body(self):
        req_data = []
        names = list(set(self.params().keys()) - set(['org_id', 'username']))

        for name in names:
            value = self.get_param(name)
            if value is not None:
                req_data.append('{}={}'.format(name, value))

        return '&'.join(req_data)



from aeronaut.request.cloud.v0_9.request import Request


class DoesImageNameExist(Request):

    def params(self):
        return {
            'image_name': {
                'required': True
            },

            'location': {
                'required': True
            },

            'org_id': {
                'required': True
            }
        }

    def http_method(self):
        return 'post'

    def url(self):
        t = '{base_url}/{org_id}/image/nameExists'

        return t.format(base_url=self.base_url,
                        org_id=self.get_param('org_id'))

    def body(self):
        template = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
            <ImageNameExists xmlns="http://oec.api.opsource.net/schemas/server">
                <location>{location}</location>
                <imageName>{image_name}</imageName>
            </ImageNameExists>
        """  # NOQA

        return template.format(image_name=self.get_param('image_name'),
                               location=self.get_param('location'))



from aeronaut.resource.cloud.disk import AdditionalDiskList, DiskList
from aeronaut.resource.cloud.machine_status import MachineStatusList
from aeronaut.resource.cloud.os import OperatingSystem
from aeronaut.resource.cloud.resource import Resource, ResourceList
from aeronaut.resource.cloud.software_label import SoftwareLabelList
from aeronaut.resource.cloud.action_status import ActionStatus


class Image(Resource):

    def _root_(self):
        return "image"

    def _members_(self):
        return {
            "cpu_count": {
                "xpath": "./*[local-name()='cpuCount']",
                "type": int},

            "created": {
                "xpath": "./*[local-name()='created']"},

            "description": {
                "xpath": "./*[local-name()='description']"},

            "disks": {
                "xpath": ".",
                "type": DiskList},

            "id": {
                "xpath": "./@*[local-name()='id']"},

            "location": {
                "xpath": "./@*[local-name()='location']"},

            "memory_mb": {
                "xpath": "./*[local-name()='memoryMb']",
                "type": int},

            "name": {
                "xpath": "./*[local-name()='name']"},

            "os": {
                "xpath": "./*[local-name()='operatingSystem']",
                "type": OperatingSystem},

            "software_labels": {
                "xpath": ".",
                "type": SoftwareLabelList},

            "source": {
                "xpath": "./*[local-name()='source']",
                "type": ImageSource},

            "state": {
                "xpath": "./*[local-name()='state']"}
        }


class ImageList(ResourceList):

    def _root_(self):
        return 'ImagesWithDiskSpeed'

    def _items_(self):
        return {
            "xpath": "./*[local-name()='image']",
            "type": Image
        }


class ImageNameExists(Resource):

    def _root_(self):
        return "Exists"

    def _members_(self):
        return {
            "is_true": {
                "xpath": ".",
                "type": bool}
        }


class ImageSource(Resource):

    def _root_(self):
        return "source"

    def _members_(self):
        return {
            "artifacts": {
                "xpath": ".",
                "type": ImageSourceArtifactList},

            "type": {
                "xpath": "./@*[local-name()='type']"}
        }


class ImageSourceArtifact(Resource):

    def _root_(self):
        return "artifact"

    def _members_(self):
        return {
            "date": {
                "xpath": "./@*[local-name()='date']"},

            "type": {
                "xpath": "./@*[local-name()='type']"},

            "value": {
                "xpath": "./@*[local-name()='value']"}
        }


class ImageSourceArtifactList(ResourceList):

    def _root_(self):
        return 'source'

    def _items_(self):
        return {
            "xpath": "./*[local-name()='artifact']",
            "type": ImageSourceArtifact
        }


class ServerImageOperatingSystem(Resource):

    def _root_(self):
        return "operatingSystem"

    def _members_(self):
        return {
            "id": {
                "xpath": "./*[local-name()='id']"},

            "display_name": {
                "xpath": "./*[local-name()='displayName']"},

            "type": {
                "xpath": "./*[local-name()='type']"},
        }


class ServerImage(Image):

    def _root_(self):
        return "ServerImageWithState"

    def _members_(self):
        base = super(ServerImage, self)._members_()

        override = {
            "deployed_time": {
                "xpath": "./*[local-name()='deployedTime']"},

            "disks": {
                "xpath": ".",
                "type": AdditionalDiskList},

            "id": {
                "xpath": "./*[local-name()='id']"},

            "location": {
                "xpath": "./*[local-name()='location']"},

            "os": {
                "xpath": "./*[local-name()='operatingSystem']",
                "type": ServerImageOperatingSystem},

            "os_storage_gb": {
                "xpath": "./*[local-name()='osStorageGb']",
                "type": int},

            "machine_status": {
                "xpath": ".",
                "type": MachineStatusList},

            "status": {
                "xpath": "./*[local-name()='status']",
                "type": ActionStatus}
        }

        # NOTE: The order of dictionaries below is important
        return dict(base.items() + override.items())



from aeronaut.request.cloud.v0_9.request import Request


class ServerActionRequest(Request):
    """Base request to be inherited by server action requests
    """

    def params(self):
        return {
            'server_id': {
                'required': True
            },

            'org_id': {
                'required': True
            }
        }

    def http_method(self):
        return 'get'

    def url(self, action):
        template = "{base_url}/{org_id}/server/{server_id}?{action}"

        return template.format(base_url=self.base_url,
                               org_id=self.get_param('org_id'),
                               server_id=self.get_param('server_id'),
                               action=action)



from aeronaut.resource.cloud.resource import Resource, ResourceList, Status


class AclRule(Resource):

    def _root_(self):
        return 'AclRule'

    def _members_(self):
        return {
            "action": {
                "xpath": "./*[local-name()='action']"},

            "dest_ip_range": {
                "xpath": "./*[local-name()='destinationIpRange']",
                "type": AclRuleDestinationIpRange},

            "id": {
                "xpath": "./*[local-name()='id']"},

            "name": {
                "xpath": "./*[local-name()='name']"},

            "port_range": {
                "xpath": "./*[local-name()='portRange']",
                "type": AclRulePortRange},

            "position": {
                "xpath": "./*[local-name()='position']",
                "type": int},

            "protocol": {
                "xpath": "./*[local-name()='protocol']"},

            "status": {
                "xpath": "./*[local-name()='status']"},

            "source_ip_range": {
                "xpath": "./*[local-name()='sourceIpRange']",
                "type": AclRuleSourceIpRange},

            "type": {
                "xpath": "./*[local-name()='type']"}
        }


class AclRuleIpRange(Resource):

    def _members_(self):
        return {
            "ip_address": {
                "xpath": "./*[local-name()='ipAddress']"},

            "netmask": {
                "xpath": "./*[local-name()='netmask']"}
        }


class AclRuleDestinationIpRange(AclRuleIpRange):

    def _root_(self):
        return "destinationIpRange"


class AclRuleList(ResourceList):

    def _root_(self):
        return 'AclRuleList'

    def _items_(self):
        return {
            "xpath": "./*[local-name()='AclRule']",
            "type": AclRule
        }


class AclRulePortRange(Resource):

    def _root_(self):
        return "portRange"

    def _members_(self):
        return {
            "type": {
                "xpath": "./*[local-name()='type']"},

            "port1": {
                "xpath": "./*[local-name()='port1']"},

            "port2": {
                "xpath": "./*[local-name()='port2']"}
        }


class AclRuleSourceIpRange(AclRuleIpRange):

    def _root_(self):
        return "sourceIpRange"


class CreateAclRuleStatus(Status):
    pass


class DeleteAclRuleStatus(Status):
    pass



from lxml import etree, objectify
import math


# =========
# EXCEPTION
# =========

class MissingResourceMembersDeclaration(Exception):

    def __init__(self, obj):
        message = 'class {} does not have a _members_ method' \
                  .format(obj.__class__.__name__)
        super(MissingResourceMembersDeclaration, self).__init__(message)


class MissingRootElementDeclarationError(Exception):

    def __init__(self, obj):
        message = 'class {} does not have a _root_ method' \
                  .format(obj.__class__.__name__)
        super(MissingRootElementDeclarationError, self).__init__(message)


class RootElementNotFoundError(Exception):

    def __init__(self, tags, xml):
        xmlstr = etree.tostring(xml)
        message = "Unable to find element '{}' in {}" \
                  .format(', '.join(tags), xmlstr)
        super(RootElementNotFoundError, self).__init__(message)


# ==============
# RESOURCE CLASS
# ==============

class Resource(object):

    def __init__(self, xml=None, **kwargs):
        xml_or_str = xml

        if not hasattr(self, '_members_') or self._members_() is None:
            raise MissingResourceMembersDeclaration(self)

        self.__ensure_members()

        # Initialize member values from the provided xml document
        if isinstance(xml_or_str, objectify.ObjectifiedElement) \
                or (isinstance(xml_or_str, str) and len(xml_or_str) > 0):

            if isinstance(xml_or_str, str):
                xml = objectify.fromstring(xml_or_str)
            else:
                xml = xml_or_str

            if not hasattr(self, '_root_'):
                raise MissingRootElementDeclarationError(self)

            root_tags = self._root_()

            if not isinstance(root_tags, list):
                root_tags = [root_tags]

            if xml.tag.split('}')[-1] not in root_tags:
                raise RootElementNotFoundError(root_tags, xml)

            root = xml

            attrs = self._members_()

            for key in attrs.keys():
                attr = attrs[key]

                if 'type' not in attr:
                    klass = str
                else:
                    klass = attr['type']

                if 'xpath' in attr:
                    result = root.xpath(attr['xpath'])
                    if result:
                        # Note that this returns a 'smart' string that knows
                        # about its orgiins and also has an xpath() method.
                        # See: http://lxml.de/tutorial.html
                        value = result[0]
                    else:
                        value = ''
                else:
                    value = ''

                if klass == bool:
                    # bool always evaluates strings as True (including the
                    # string 'false') so we have to treat boolean fields
                    # differently
                    value = value in ['True', True, 'true']
                elif klass == "auto":
                    value = self.__autocast(value)
                else:
                    try:
                        value = klass(value)
                    except ValueError:
                        value = None

                self.__init_property(key, value)

            # If the xml object has child elements named 'property', those will
            # be available via the [] accessor where the property name is the
            # key and the property value is the value.
            self.__items = {
                p.get('name'): self.__autocast(p.get('value'))
                for p in root.xpath("./*[local-name()='property']")
            }

        # Prevent random attributes from being added going forward
        self.__freeze()

    # ==============
    # PUBLIC METHODS
    # ==============

    def to_dict(self, only_dirty=False):
        """
        Returns the resource's members and their values as a dictionary.

        Args:
            only_dirty (bool): When True, returns only the members that changed
        """
        if only_dirty:
            properties = self.__dirty_properties
        else:
            properties = self.__all_properties

        items = {}

        for name in properties:
            items[name] = self.__get_property(name)

        return items

    # =============
    # MAGIC METHODS
    # =============

    def __getitem__(self, key):
        return self.__items[key]

    __isfrozen = False

    def __setattr__(self, key, value):
        if self.__isfrozen and not hasattr(self, key):
            raise AttributeError("Unknown attribute {}".format(key))
        object.__setattr__(self, key, value)

    # ===============
    # PRIVATE METHODS
    # ===============

    def __autocast(self, value):
        try:
            return int(value)
        except (ValueError, TypeError):
            pass

        try:
            return float(value)
        except (ValueError, TypeError):
            pass

        return str(value)

    def __add_property(self, name):
        """
        Adapted from http://goo.gl/si4kmX

        This dynamically creates the Python properties from the attributes
        defined in _members_(). We do it this way so:

            1) the attributes are visible when inspecting the object via
               dir(object_instance) and

            2) more importantly, so we can freeze the instance after
               initialization, preventing accidental attribute creation due
               to misspellings.
        """
        getter = lambda self: self.__get_property(name)
        setter = lambda self, value: self.__set_property(name, value)
        setattr(self.__class__, name, property(getter, setter))

    def __ensure_members(self):
        attrs = self._members_()

        self.__all_properties = attrs.keys()
        self.__dirty_properties = []

        for key in attrs.keys():
            # This creates a getter and setter in the class if it doesn't exist
            if not hasattr(self.__class__, key):
                self.__add_property(key)

            # This creates an attribute in the instance
            self.__init_property(key, None)

    def __freeze(self):
        self.__isfrozen = True

    def __get_property(self, name):
        """
        Returns the value of the given property
        """
        return getattr(self, '__property_' + name)

    def __init_property(self, name, value):
        """
        Initialize the property. This is used by __init__ for setting the value
        of the given property without marking the property "dirty"
        """
        setattr(self, '__property_' + name, value)

    def __set_property(self, name, value):
        """
        Sets the value of the given property and marks it as "dirty"
        """
        if name not in self.__dirty_properties:
            self.__dirty_properties.append(name)
        setattr(self, '__property_' + name, value)


# ===================
# RESOURCE LIST CLASS
# ===================

class ResourceList(object):

    def __init__(self, xml=None, resources=None):
        if isinstance(xml, objectify.ObjectifiedElement) \
                or (isinstance(xml, str) and len(xml) > 0):

            if isinstance(xml, str):
                xml = objectify.fromstring(xml)

            if not hasattr(self, '_root_'):
                raise MissingRootElementDeclarationError(self)

            root_tags = self._root_()

            if not isinstance(root_tags, list):
                root_tags = [root_tags]

            if xml.tag.split('}')[-1] not in root_tags:
                raise RootElementNotFoundError(root_tags, xml)

            self.__root_node = xml

            xpath = self._items_()['xpath']

            if 'type' not in self._items_():
                klass = str
            else:
                klass = self._items_()['type']

            self.__items = [klass(item) for item in xml.xpath(xpath)]

    def __getitem__(self, index):
        return self.__items[index]

    def __len__(self):
        return len(self.__items)

    @property
    def page_number(self):
        return self._get_paging_value_or_none('pageNumber')

    @property
    def page_size(self):
        return self._get_paging_value_or_none('pageSize')

    @property
    def total_count(self):
        return self._get_paging_value_or_none('totalCount')

    @property
    def total_pages(self):
        total_count = self._get_paging_value_or_none('totalCount')
        page_size = self._get_paging_value_or_none('pageSize')

        if total_count:
            return int(math.ceil(float(total_count) / page_size))
        else:
            return None

    def _get_paging_value_or_none(self, name):
        v = self.__root_node.xpath("./@{}".format(name))

        if v:
            return int(v[0])
        else:
            return None


# ============
# STATUS CLASS
# ============

class StatusAdditionalInfo(Resource):

    def _root_(self):
        return "additionalInformation"

    def _members_(self):
        return {
            "name": {
                "xpath": "./@*[local-name()='name']"},

            "value": {
                "xpath": "./*[local-name()='value']"},
        }


class StatusAdditionalInfoList(ResourceList):

    def _root_(self):
        return 'Status'

    def _items_(self):
        return {
            "xpath": "./*[local-name()='additionalInformation']",
            "type": StatusAdditionalInfo
        }


class Status(Resource):

    def _root_(self):
        return "Status"

    def _members_(self):
        return {
            "additional_info": {
                "xpath": ".",
                "type": StatusAdditionalInfoList},

            "operation": {
                "xpath": "./*[local-name()='operation']"},

            "result": {
                "xpath": "./*[local-name()='result']"},

            "result_detail": {
                "xpath": "./*[local-name()='resultDetail']"},

            "result_code": {
                "xpath": "./*[local-name()='resultCode']"}
        }

    @property
    def is_success(self):
        return self.result_code == 'REASON_0'



from aeronaut.request.cloud.v0_9.request import Request


class CreateNetwork(Request):

    def params(self):
        return {
            'description': {
                'required': False,
                'default': ''
            },
            'location': {
                'required': True
            },
            'name': {
                'required': True
            },
            'org_id': {
                'required': True
            }
        }

    def http_method(self):
        return 'post'

    def url(self):
        template = '{base_url}/{org_id}/networkWithLocation'

        return template.format(base_url=self.base_url,
                               org_id=self.get_param('org_id'))

    def body(self):
        template = """
        <NewNetworkWithLocation xmlns="http://oec.api.opsource.net/schemas/network">
            <name>{name}</name>
            <description>{description}</description>
            <location>{location}</location>
        </NewNetworkWithLocation>
        """  # NOQA

        return template.format(name=self.get_param('name'),
                               description=self.get_param('description'),
                               location=self.get_param('location'))



from aeronaut.request.cloud.v0_9.request import Request


class GetServerImage(Request):

    def params(self):
        return {
            'org_id': {
                'required': True
            },

            'image_id': {
                'required': True
            }
        }

    def http_method(self):
        return 'get'

    def url(self):
        t = '{base_url}/{org_id}/image/{image_id}'

        return t.format(base_url=self.base_url,
                        org_id=self.get_param('org_id'),
                        image_id=self.get_param('image_id'))



from aircraft.request import Request as BaseRequest


class Request(BaseRequest):

    @property
    def base_url(self):
        if hasattr(self, '_base_url'):
            return '{}/oec/0.9'.format(self._base_url)



from aeronaut.request.cloud.v0_9.request import Request


class ListImages(Request):

    def with_paging(self):
        return True

    def fields(self):
        return {
            'created': {
                'filter': True,
                'sort': True
            },

            'image_id': {
                'filter': True,
                'sort': True,
                'query_key': 'id'
            },

            'location': {
                'filter': True,
                'sort': True
            },

            'name': {
                'filter': True,
                'sort': True
            },

            'os_family': {
                'filter': True,
                'sort': True,
                'query_key': 'operatingSystemFamily'
            },

            'os_id': {
                'filter': True,
                'sort': True,
                'query_key': 'operatingSystemId'
            },

            'state': {
                'filter': True,
                'sort': True
            }
        }

    def params(self):
        return {
            'base_or_org_id': {
                'required': True
            }
        }

    def http_method(self):
        return 'get'

    def url(self):
        t = '{base_url}/{base_or_org_id}/imageWithDiskSpeed?'

        return t.format(base_url=self.base_url,
                        base_or_org_id=self.get_param('base_or_org_id'))



from aeronaut.request.cloud.v0_9.request import Request


class DeployServer(Request):

    def params(self):
        return {
            'admin_password': {
                'required': False,
                'default': None
            },

            'description': {
                'required': False,
                'default': ''
            },

            'disks': {
                'required': False,
                'default': []
            },

            'image_id': {
                'required': True
            },

            'name': {
                'required': True
            },

            'network_id': {
                'required': False,
                'default': None
            },

            'org_id': {
                'required': True
            },

            'private_ip': {
                'required': False,
                'default': None
            },

            'start': {
                'required': False,
                'default': True
            }
        }

    def http_method(self):
        return 'post'

    def url(self):
        template = '{base_url}/{org_id}/deployServer'

        return template.format(base_url=self.base_url,
                               org_id=self.get_param('org_id'))

    def body(self):
        template = """
            <DeployServer xmlns="http://oec.api.opsource.net/schemas/server">
                <name>{name}</name>
                <description>{description}</description>
                <imageId>{image_id}</imageId>
                <start>false</start>"""

        if self.has_param('admin_password'):
            template += """
                <administratorPassword>{admin_password}</administratorPassword>"""  # NOQA

        if self.has_param('network_id'):
            template += """
                <networkId>{network_id}</networkId>"""

        if self.has_param('private_ip'):
            template += """
                <privateIp>{private_ip}</privateIp>"""

        for speed in self.get_param('disks'):
            template += """
                <disk scsiId="{scsi_id}" speed="{speed}"/>""" \
                .format(scsi_id=speed['scsi_id'],
                        speed=speed['speed'])

        template += """
            </DeployServer>"""

        return template.format(admin_password=self.get_param('admin_password'),
                               name=self.get_param('name'),
                               description=self.get_param('description'),
                               image_id=self.get_param('image_id'),
                               location=self.get_param('location'),
                               network_id=self.get_param('network_id'),
                               private_ip=self.get_param('private_ip'),
                               start=str(self.get_param('start')).lower())



from aeronaut.resource.cloud.resource import Resource, ResourceList


class MachineStatus(Resource):

    def _root_(self):
        return "machineStatus"

    def _members_(self):
        return {
            "name": {
                "xpath": "./@*[local-name()='name']"},

            "value": {
                "xpath": "./*[local-name()='value']",
                "type": "auto"}
        }


class MachineStatusList(ResourceList):

    def _root_(self):
        return [
            "server",
            "ServerImageWithState"
        ]

    def _items_(self):
        return {
            "xpath": "./*[local-name()='machineStatus']",
            "type": MachineStatus
        }



from aeronaut.request.cloud.v0_9.request import Request


class ListDataCenters(Request):

    def with_paging(self):
        return True

    def fields(self):
        return {
            'location': {
                'filter': True,
                'sort': True
            }
        }

    def params(self):
        return {
            'org_id': {
                'required': True
            }
        }

    def http_method(self):
        return 'get'

    def url(self):
        return '{base_url}/{org_id}/datacenterWithMaintenanceStatus?' \
               .format(base_url=self.base_url,
                       org_id=self.get_param('org_id'))



from aircraft.params import Params


# ==========
# EXCEPTIONS
# ==========

class UnknownFilterKeyError(Exception):

    def __init__(self, invalid_key, valid_keys):
        message = "'{invalid_key}' is invalid. Valid filters are " \
                  "{valid_keys}.".format(invalid_key=invalid_key,
                                         valid_keys=", ".join(valid_keys))
        super(UnknownFilterKeyError, self).__init__(message)


class UnknownSortKeyError(Exception):

    def __init__(self, invalid_key, valid_keys):
        message = "'{invalid_key}' is invalid. Valid sort keys are " \
                  "{valid_keys}.".format(invalid_key=invalid_key,
                                         valid_keys=", ".join(valid_keys))
        super(UnknownSortKeyError, self).__init__(message)


# ==================
# BASE REQUEST CLASS
# ==================

class Request(object):

    def __init__(self, auth_data=None, base_url=None, params={}):
        self._auth_data = auth_data
        self._base_url = base_url

        params_def = self._build_params_def()

        if params_def:
            self.__params__ = Params(params_def, params)

    # =================
    # PUBLIC PROPERTIES
    # =================

    @property
    def base_url(self):
        return self._base_url

    # ==============
    # PUBLIC METHODS
    # ==============

    def basic_auth(self):
        return None

    def body(self):
        return None

    def build_url(self):
        url = self.url()

        if self.with_paging():
            query = []
            page_number = self.get_param('page_number')
            page_size = self.get_param('page_size')

            if page_number:
                query.append("pageNumber={}".format(page_number))

            if page_size:
                query.append("pageSize={}".format(page_size))

            if '?' not in url and query:
                url += "?"

            if query:
                url += "&".join(query)

        # We expect filters here to be an array of 3-element arrays
        # with each element being a string:
        #       [[<field name>, <operator>, <value>], [<f>,<o>,<v>], ...]
        filters = self.get_param('filters')

        # However...

        # We can accommodate the scenario where filters is just
        # a 3-element array:
        #       [<field name>, <operator>, <value>]
        #
        # and this block will just convert that into an array of
        # 3-element arrays.
        if isinstance(filters, list) \
                and len(filters) == 3 \
                and all([isinstance(e, str) for e in filters]):
            filters = [filters]

        if filters:
            query = "&".join([
                "{key}{operator}{value}".format(
                    key=self._to_filter_query_key(f[0]),
                    operator=self._to_operator(f[1].upper()),
                    value=f[2])
                for f in filters
            ])

            if "?" not in url and query:
                url += "?"

            if query and url[-1] != "?":
                url += "&"

            url += query

        sorters = self.get_param("sort")

        if sorters:
            query = ",".join([
                "{key}.{direction}".format(
                    key=self._to_sort_query_key(s),
                    direction=self._to_sort_dir(s))
                for s in sorters
            ])

            if "?" not in url and query:
                url += "?"

            if query and url[-1] != "?":
                url += "&"

            url += "orderBy=" + query

        return url

    def get_param(self, name):
        if self.has_param(name):
            return getattr(self.__params__, name)
        else:
            return None

    def has_param(self, name):
        return hasattr(self, '__params__') and \
            hasattr(self.__params__, name) and \
            getattr(self.__params__, name) is not None

    def headers(self):
        return {}

    def http_method(self):
        return None

    def url(self):
        return None

    def with_paging(self):
        return False

    # ===============
    # PRIVATE METHODS
    # ===============

    def _build_params_def(self):
        if hasattr(self, 'params'):
            params_def = self.params()
        else:
            params_def = {}

        if self.with_paging():
            for param in ['page_number', 'page_size']:
                params_def[param] = {
                    'required': False,
                    'default': None
                }

        if hasattr(self, 'fields'):
            fields = self.fields()
            filterable = any([
                'filter' in v and v['filter']
                for k, v in fields.items()
            ])
            sortable = any([
                'sort' in v and v['sort']
                for k, v in fields.items()
            ])

            if filterable:
                params_def['filters'] = {
                    'required': False,
                    'default': []
                }
            if sortable:
                params_def['sort'] = {
                    'required': False,
                    'default': []
                }

        return params_def

    def _to_filter_query_key(self, string):
        fields = self.fields()
        valid_keys = [key for key, d
                      in fields.items()
                      if 'filter' in d and d['filter']]

        if string not in valid_keys:
            raise UnknownFilterKeyError(string, valid_keys)

        return self._to_query_key(string, fields)

    def _to_operator(self, string):
        if string.upper() in ["EQ", "EQUALS", "=", "=="]:
            return "="
        else:
            return ".{}=".format(string.upper())

    def _to_query_key(self, string, fields):
        if 'query_key' in fields[string]:
            return fields[string]['query_key']
        elif '_' in string and len(string.split('_')) > 1:
            parts = string.split('_')
            return ''.join(parts[0:1] + [s.capitalize() for s in parts[1:]])
        else:
            return string

    def _to_sort_dir(self, string):
        parts = string.split(" ")

        if len(parts) < 2:
            return "ASCENDING"
        elif parts[1].upper() in ["ASC", "ASCENDING"]:
            return "ASCENDING"
        else:
            return "DESCENDING"

    def _to_sort_query_key(self, string):
        string = string.split(" ")[0]
        fields = self.fields()
        valid_keys = [key for key, d
                      in fields.items()
                      if 'sort' in d and d['sort']]

        if string not in valid_keys:
            raise UnknownSortKeyError(string, valid_keys)

        return self._to_query_key(string, fields)



from aeronaut.resource.cloud.action_status import ActionStatus
from aeronaut.resource.cloud.disk import DiskList
from aeronaut.resource.cloud.machine_status import MachineStatusList
from aeronaut.resource.cloud.os import OperatingSystem
from aeronaut.resource.cloud.resource import Resource, ResourceList, Status
from aeronaut.resource.cloud.software_label import SoftwareLabelList


class AddStorageToServerStatus(Status):
    pass


class Backup(Resource):

    def _root_(self):
        return "backup"

    def _members_(self):
        return {
            "asset_id": {
                "xpath": "./@*[local-name()='assetId']"},

            "service_plan": {
                "xpath": "./@*[local-name()='servicePlan']"},

            "state": {
                "xpath": "./@*[local-name()='state']"}
        }


class CleanFailedServerDeploymentStatus(Status):
    pass


class DeleteServerStatus(Status):
    pass


class DeployServerStatus(Status):
    pass


class ModifyServerStatus(Status):
    pass


class PoweroffServerStatus(Status):
    pass


class RebootServerStatus(Status):
    pass


class ResetServerStatus(Status):
    pass


class Server(Resource):

    def _root_(self):
        return "server"

    def _members_(self):
        return {
            "backup": {
                "xpath": "./*[local-name()='backup']",
                "type": Backup},

            "cpu_count": {
                "xpath": "./*[local-name()='cpuCount']",
                "type": int},

            "created": {
                "xpath": "./*[local-name()='created']"},

            "description": {
                "xpath": "./*[local-name()='description']"},

            "disks": {
                "xpath": ".",
                "type": DiskList},

            "id": {
                "xpath": "./@*[local-name()='id']"},

            "is_deployed": {
                "xpath": "./*[local-name()='isDeployed']",
                "type": bool},

            "is_started": {
                "xpath": "./*[local-name()='isStarted']",
                "type": bool},

            "location": {
                "xpath": "./@*[local-name()='location']"},

            "machine_name": {
                "xpath": "./*[local-name()='machineName']"},

            "machine_status": {
                "xpath": ".",
                "type": MachineStatusList},

            "memory_mb": {
                "xpath": "./*[local-name()='memoryMb']",
                "type": int},

            "name": {
                "xpath": "./*[local-name()='name']"},

            "network_id": {
                "xpath": "./*[local-name()='networkId']"},

            "os": {
                "xpath": "./*[local-name()='operatingSystem']",
                "type": OperatingSystem},

            "public_ip": {
                "xpath": "./*[local-name()='publicIp']"},

            "private_ip": {
                "xpath": "./*[local-name()='privateIp']"},

            "software_labels": {
                "xpath": ".",
                "type": SoftwareLabelList},

            "source_image_id": {
                "xpath": "./*[local-name()='sourceImageId']"},

            "state": {
                "xpath": "./*[local-name()='state']"},

            "status": {
                "xpath": "./*[local-name()='status']",
                "type": ActionStatus}
        }


class ServerList(ResourceList):

    def _root_(self):
        return "ServersWithBackup"

    def _items_(self):
        return {
            "xpath": "./*[local-name()='server']",
            "type": Server
        }


class ShutdownServerStatus(Status):
    pass


class StartServerStatus(Status):
    pass



#! /usr/bin/env python

# Some more useful tips to adopt here http://goo.gl/LbBM3D

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand  # NOQA
import sys


class Tox(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import tox
        errcode = tox.cmdline(self.test_args)
        sys.exit(errcode)


setup(name='Aeronaut',
      version='0.1.0a3',
      description='Aeronaut: A client library for DiData Cloud',
      author='Nexus IS Dev Team',
      author_email='dev@nexusis.com',
      packages=find_packages(),
      test_suite='aeronaut.test',
      tests_require=['tox'],
      cmdclass={'test': Tox},
      install_requires=[
          'PyYAML >=3.11',
          'requests >=2.5.1',
          'lxml >=3.4.2'
      ])



from aeronaut.resource.cloud.resource import Resource


class OperatingSystem(Resource):

    def _root_(self):
        return "operatingSystem"

    def _members_(self):
        return {
            "id": {
                "xpath": "./@*[local-name()='id']"},

            "display_name": {
                "xpath": "./@*[local-name()='displayName']"},

            "type": {
                "xpath": "./@*[local-name()='type']"},
        }





# ==========
# EXCEPTIONS
# ==========

class UnknownContentTypeError(Exception):

    def __init__(self, content_type):
        message = 'Unknown content-type: {}' \
                  .format(content_type)
        super(UnknownContentTypeError, self).__init__(message)


# ==============
# RESPONSE CLASS
# ==============

class Response(object):

    def __init__(self, response, request):
        self._response = response
        self._request = request

    # =============
    # MAGIC METHODS
    # =============

    def __getattr__(self, name):
        return getattr(self._response, name)

    # =================
    # PUBLIC PROPERTIES
    # =================

    @property
    def content_type(self):
        return self._response.headers['content-type'].split(';')[0]

    @property
    def body(self):
        return self._response.content



from aeronaut.resource.cloud.resource import Resource, ResourceList


class Account(Resource):

    def _root_(self):
        return 'Account'

    def _members_(self):
        return {
            "department": {
                "xpath": "./*[local-name()='department']"},

            "email": {
                "xpath": "./*[local-name()='emailAddress']"},

            "first_name": {
                "xpath": "./*[local-name()='firstName']"},

            "full_name": {
                "xpath": "./*[local-name()='fullName']"},

            "last_name": {
                "xpath": "./*[local-name()='lastName']"},

            "org_id": {
                "xpath": "./*[local-name()='orgId']"},

            "password": {
                "writeonly": True},

            "roles": {
                "xpath": "./*[local-name()='roles']",
                "type": RoleList},

            "username": {
                "xpath": "./*[local-name()='userName']"}
        }


class RoleList(ResourceList):

    def _root_(self):
        return 'roles'

    def _items_(self):
        return {
            "xpath": "./*[local-name()='role']",
            "type": Role
        }


class Role(Resource):

    def _root_(self):
        return 'role'

    def _members_(self):
        return {
            "name": {"xpath": "./*[local-name()='name']"}
        }



class ValueObject(object):

    def __eq__(self, other):
        return self._attrs == other._attrs

    def __getattr__(self, name):
        return self._attrs[name]

    def __hash__(self):
        return hash(frozenset(self._attrs.items()))

    def __ne__(self, other):
        return not self.__eq__(other)



from aeronaut.request.cloud.v0_9.server_action import ServerActionRequest


class CleanFailedServerDeployment(ServerActionRequest):

    def url(self):
        return super(CleanFailedServerDeployment, self).url('clean')



from aeronaut.request.cloud.v0_9.request import Request


class ListNetworks(Request):

    def params(self):
        return {
            'location': {
                'required': False,
                'default': None
            },
            'org_id': {
                'required': True
            }
        }

    def http_method(self):
        return 'get'

    def url(self):
        template = '{base_url}/{org_id}/networkWithLocation'

        if self.has_param('location'):
            template += '/{location}'

        return template.format(base_url=self.base_url,
                               location=self.get_param('location'),
                               org_id=self.get_param('org_id'))
